
/*
libfawk - A function-only AWK dialect - compacted, single-source-file version
(for a human readable source code please visit the project page)

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS AS IS AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#ifndef fawk_malloc
#define fawk_malloc(ctx,size) malloc(size)
#define fawk_calloc(ctx,n,m) calloc(n, m)
#define fawk_realloc(ctx,ptr,size) realloc(ptr, size)
#define fawk_free(ctx,ptr) free(ptr)
#endif
#ifndef fawk_assert
#include <assert.h>
#define fawk_assert assert
#endif
#ifndef fawk_no_math
#include <math.h>
#define fawk_strtod(str) strtod(str, NULL)
#define fawk_fmod(x,y) fmod(x, y)
#define FAWK_NUM_PRINTF_FMT "%g"
#endif
extern FILE *fdopen(int fd, const char *mode);
static void stdio_uninit_cb(fawk_pkg_t *pkg, fawk_ctx_t *ctx)
{
 fawk_htpp_t *files = pkg->data[0].ptr;
 fawk_htpp_entry_t *e;
 for(e = fawk_htpp_first(files); e != NULL; e = fawk_htpp_next(files, e)) {
  fclose(e->value);
  fawk_free(ctx, e->key);
 }
 fawk_htpp_uninit(files);
 fawk_free(ctx, files);
}
static fawk_pkg_t *get_pkg(fawk_ctx_t *ctx)
{
 fawk_pkg_t *pkg;
 for(pkg = ctx->pkg_head; pkg != NULL; pkg = pkg->next)
  if (pkg->uninit_cb == stdio_uninit_cb)
   return pkg;
 return NULL;
}
static FILE *fawk_fopen(fawk_ctx_t *ctx, const char *fn, const char *mode)
{
 FILE *f;
 fawk_htpp_t *files;
 fawk_pkg_t *pkg;
 if (*fn == '/') {
  if (strcmp(fn, "/dev/stdin") == 0) return stdin;
  if (strcmp(fn, "/dev/stdout") == 0) return stdout;
  if (strcmp(fn, "/dev/stderr") == 0) return stderr;
 }
 pkg = get_pkg(ctx);
 if (pkg == NULL) return NULL;
 files = pkg->data[0].ptr;
 if (files == NULL) return NULL;
 f = fawk_htpp_get(files, fn);
 if (f != NULL) return f;
 if ((*fn == '/') && (strncmp(fn, "/dev/fd/", 8) == 0))
  f = fdopen(atoi(fn+8), mode);
 else
  f = fopen(fn, mode);
 if (f == NULL) return NULL;
 fawk_htpp_set(files, fawk_strdup(ctx, fn), f);
 return f;
}
static void fawk_fclose(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
 fawk_htpp_entry_t *e;
 fawk_pkg_t *pkg;
 fawk_htpp_t *files;
 if ((argc != 1) || (FAWK_CFUNC_ARG(0)->type != FAWK_STR)) return;
 pkg = get_pkg(ctx);
 if (pkg == NULL) return;
 files = pkg->data[0].ptr;
 for(e = fawk_htpp_first(files); e != NULL; e = fawk_htpp_next(files, e)) {
  if (strcmp(e->key, FAWK_CFUNC_ARG(0)->data.str->str) == 0) {
   fclose(e->value);
   fawk_htpp_pop(files, e->key);
   break;
  }
 }
}
#define APPEND(c) \
do { \
 if (s == NULL) { \
  s = fawk_malloc(ctx, sizeof(fawk_str_t) + 64); \
  s->used = 0; \
  s->alloced = 64; \
 } \
 else if (s->used >= s->alloced) { \
  if (s->alloced < 2048) s->alloced *= 2; \
  else s->alloced += 1024; \
  s = fawk_realloc(ctx, s, sizeof(fawk_str_t) + s->alloced+1); \
 } \
 s->str[s->used++] = c; \
} while(0)
static void fawk_fgetline(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
 fawk_str_t *s = NULL;
 FILE *f;
 if ((argc < 1) || (argc > 2) || (FAWK_CFUNC_ARG(0)->type != FAWK_STR))
  goto ret_eof;
 if ((argc > 1) && (FAWK_CFUNC_ARG(1)->type != FAWK_STR))
  goto ret_eof;
 f = fawk_fopen(ctx, FAWK_CFUNC_ARG(0)->data.str->str, "r");
 if ((f == NULL) || (feof(f)))
  goto ret_eof;
 for(;;) {
  int c =fgetc(f);
  if (c == EOF) break;
  if (argc > 1) {
   if (strchr(FAWK_CFUNC_ARG(1)->data.str->str, c) != NULL) break;
  }
  else if (c == '\n')
   break;
  APPEND(c);
 }
 if (s == NULL)
  return;
 s->str[s->used++] = '\0';
 s->refco = 1;
 retval->type = FAWK_STR;
 retval->data.str = s;
 return;
 ret_eof:;
 {
  fawk_pkg_t *pkg = get_pkg(ctx);
  if (pkg == NULL) return;
  memcpy(retval, pkg->data[1].ptr, sizeof(fawk_cell_t));
  retval->data.str->refco++;
 }
}
static void fawk_fprintf(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
 FILE *f;
 const char *fmt;
 int curr = 2;
 fawk_cell_t *ac;
 if ((argc < 2) || (FAWK_CFUNC_ARG(0)->type != FAWK_STR) || (FAWK_CFUNC_ARG(1)->type != FAWK_STR)) return;
 f = fawk_fopen(ctx, FAWK_CFUNC_ARG(0)->data.str->str, "w");
 if (f == NULL)
  return;
 for(fmt = FAWK_CFUNC_ARG(1)->data.str->str; *fmt != '\0'; fmt++) {
  if ((*fmt == '%') && (curr < argc)) {
   char ofmt[128];
   const char *start = fmt++;
   if ((*fmt == '-') || (*fmt == '+')) fmt++;
   while(isdigit(*fmt) || (*fmt == '.')) fmt++;
   if ((fmt - start + 1) < sizeof(ofmt)-1) {
    memcpy(ofmt, start, (fmt - start + 1));
    ofmt[fmt - start + 1] = '\0';
    ac = FAWK_CFUNC_ARG(curr);
    switch(*fmt) {
     case 's':
      if (fawk_cast_to_str(ctx, ac) == 0)
       fprintf(f, ofmt, ac->data.str->str);
      break;
     case 'd':
      if (fawk_cast_to_num(ctx, ac) == 0)
       fprintf(f, ofmt, (int)ac->data.num);
      break;
     case 'f':
      if (fawk_cast_to_num(ctx, ac) == 0)
       fprintf(f, ofmt, (double)ac->data.num);
      break;
     default:
      fprintf(f, "<invalid format %s>", ofmt);
    }
   }
   else
    fprintf(f, "<format too long>");
   curr++;
  }
  else
   fputc(*fmt, f);
 }
}
FAWK_API int fawk_stdio_init(fawk_ctx_t *ctx)
{
 fawk_pkg_t *pkg;
 fawk_htpp_t *files;
 fawk_cell_t *v;
 v = fawk_symtab_regvar(ctx, "FEOF", FAWK_SCALAR);
 v->type = FAWK_STR; v->data.str = fawk_str_new_from_literal(ctx, "\n<FEOF>\n", -1);
 if (v->data.str == NULL) return -1;
 pkg = fawk_calloc(ctx, sizeof(fawk_pkg_t), 1);
 if (pkg == NULL) return -1;
 pkg->next = ctx->pkg_head;
 ctx->pkg_head = pkg;
 pkg->uninit_cb = stdio_uninit_cb;
 pkg->data[1].ptr = v;
 fawk_htpp_init(pkg->data[0].ptr = files = fawk_malloc(ctx, sizeof(fawk_htpp_t)), strhash, strkeyeq);
 if (files == NULL) {
  free(pkg);
  return -1;
 }
 fawk_symtab_regcfunc(ctx, "fgetline", fawk_fgetline);
 fawk_symtab_regcfunc(ctx, "fclose", fawk_fclose);
 fawk_symtab_regcfunc(ctx, "fprintf", fawk_fprintf);
 return 0;
}
